/**
 * Copyright (c) 1996-2023, SR Research Ltd., All Rights Reserved
 *
 * For use by SR Research licencees only. Redistribution and use in source
 * and binary forms, with or without modification, are NOT permitted.
 *
 * Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in
 * the documentation and/or other materials provided with the distribution.
 *
 * Neither name of SR Research Ltd nor the name of contributors may be used
 * to endorse or promote products derived from this software without
 * specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS ``AS
 * IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE REGENTS OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/*******************************************************************************
 * EYELINK EXPT SUPPORT: SDL PLATFORM (Coppied from WIN32 PLATFORM)			   *
 * by Suganthan Subramaniam                                                    *
 * EYELINK II: 3 February 2003                                                 *
 * Provides text support for sdl							                   *
 *                                                                             *
 *                                                                             *
 **************************************** WARNING ******************************
 *                                                                             *
 * UNDER NO CIRCUMSTANCES SHOULD PARTS OF THESE FILES BE COPIED OR COMBINED.   *
 * This will make your code impossible to upgrade to new releases in the future*
 * and SR Research will not give tech support for reorganized code.            *
 *                                                                             *
 * This file should not be modified. If you must modify it, copy the entire    *
 * file with a new name, and change the the new file.                          *
 *******************************************************************************/


#define  _CRT_SECURE_NO_WARNINGS // for msvc, otherwise we get a lot of warnings.
#include <stdio.h>    // for vsprintf() function
#include <stdarg.h>
#include <string.h>
#include <stdlib.h>
#include <time.h>
#include <SDL/SDL.h>
#include <errno.h>

#ifdef __APPLE__ //MACOSX
#include <SDL_ttf/SDL_ttf.h>
#include <CoreFoundation/CoreFoundation.h>
#include <NSSystemDirectories.h>
#else
#include <SDL_ttf.h>
#endif
#include "fontmappings.h"
#include "sdl_text_support.h"
#ifdef _WIN32
#include <direct.h>
#include <io.h>
#if (_MSC_VER > 1200)
#define strtok_r strtok_s
#else

#endif
#define makedir(a) _mkdir(a)
#define access(a,b) _access(a,b)
#define F_OK 00
#define PATH_DELIMITER ";"


#else
#include <libgen.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>

#define min(x,y) (x>y)? y:x
#define _stricmp strcasecmp
#define PATH_DELIMITER ":"
#define makedir(a) mkdir(a,S_IRWXU)
#endif

static  TTF_Font *font = NULL;
#define FONTSIZE 30
#define MAX_CHAR_PERLINE 1024
static int marginx =0;
static int marginy =0;
static int marginw =0;
static int marginh =0;
static double linespacing = (9.0/10.0);

/*
 search path is delimited by :
 */
static char font_search_path[4096]="";
static char output_folder_path[4096]="";
static char exe_folder_path[4096]="";

static FONT_MAP fontmap[] = FONTMAPPINGS;


void set_line_spacing(double linespace)
{
	linespacing = linespace;
}
void set_margin(int left, int right, int top, int bottom)
{
  marginx = left;
  marginy =top;
  marginw =right;
  marginh =bottom;
}



/*
 * Allow to add font search folder. By default fonts are search in
 * get_resources_path() and some system folders.
 * in windows <windir>\Fonts
 * in mac, we should technically search /System/Library/Fonts, /Library/Fonts and ~/Library/Fonts
 * However, in recent macs there does not seem to be any ttf files in these foledrs.  So they are not
 * searched.
 */
void add_font_search_path(const char *path)
{
	if(strlen(font_search_path)!=0)
	{
		strcat(font_search_path,PATH_DELIMITER);
	}
	strcat(font_search_path,path);
}

int initialize_sdl_util(const char *exe_path, const char *output_folder, int create_output_folder)
{
	return initialize_folder(exe_path, output_folder, create_output_folder);
}



/*
 tries to find the font file name and returns the file name.
 search_index takes in and return the index on the map.
 For the first call assign the value of 0. On the consequtive times
 keep with the last returned value.
 if none found, searh_index is set to -1;

 eg.
 searh_index=0;
 get_font_file_name(NULL, &searh_index);
 get_font_file_name(NULL, &searh_index); --> In this call the search continues from where we left off.

 */
static char * get_font_file_name(const char *_fontname, int *searh_index, int bold )
{
	char *fontname=(char *)_fontname; //to fool the compiler from complaing
	int i =*searh_index;
        if(!fontname) fontname= FONTNAME; // default font
	for(; fontmap[i].filename && fontmap[i].font_face && fontmap[i].font_name; i++)
	{
		if(_stricmp(fontmap[i].font_name,fontname) ==0)
		{
			*searh_index=i+1;
			return fontmap[i].filename;
		}
	}
	*searh_index=-1;
	return FONTNAME; // default font
}

#if defined(_WIN32)
static const char * get_windows_font_path()
{
	static  char wd[1024] = {0};
	if(!strlen(wd))
	{
		strcpy(wd, getenv("windir"));
		strcat(wd,"/Fonts/");
	}
	return wd;
}
#endif
/*
 * returns 1 if file found in the path given
 * result contains the full path of the fille
 */
static int search_font_in_path(const char *fontname, const char *path, char * result)
{
	char fpath[2048];

	if (NULL == path)
	{
		if (NULL == fontname)
		{
			printf("Unable to search for null font in a null path.\n");
			return 0;
		}
		else
		{
			printf("Unable to search for font %s in a null path.\n", fontname);
			return 0;
		}
	}

	if(strlen(path))
		sprintf(fpath, "%s/%s",  path, fontname);
	else
		sprintf(fpath, "%s",  fontname);
	if( access( fpath, F_OK ) == 0 ) {
		strcpy(result, fpath);
		return 1;
	} else {
		return 0;
	}
}

static int search_all_font_in_path(const char *fontname,int bold, const char *path, char * result)
{
	int searh_index = 0;
	// not found search more.
	do
	{
		const char * fontfile = get_font_file_name(fontname, &searh_index, bold);
		// check current folder or abs path first.
		if (search_font_in_path(fontfile, path, result))
		{ // font file found.
			return 1;
		}
	} while (searh_index>0);
	return 0;
}

// look for font in the font search path.
static int search_for_font(char *fontname, int bold, char * result)
{

  char search_path_copy[4096] = {0};
  char* token;
  char* rest;

 	const char * searchfolder = NULL;
	if (search_all_font_in_path(fontname, bold, "", result))
		return 1;

	searchfolder = NULL;
#if defined(__APPLE__)
  // in macos we check for get_resources_path first, then any path given in the search path.
  searchfolder = get_resources_path();
#else
#if defined(_WIN32)
  searchfolder = get_windows_font_path();
#endif
#endif

  if (search_all_font_in_path(fontname, bold, searchfolder, result))
	  return 1;


// search in font search path
  rest = search_path_copy;
  strcpy(search_path_copy, font_search_path);

#if (_MSC_VER > 1200) || !defined(_WIN32)
  while ((token = strtok_r(rest, PATH_DELIMITER, &rest)))
  {
	  if (search_all_font_in_path(fontname, bold, token, result))
		  return 1;
  }
#else
// no multi-thread support
  token = strtok(rest, PATH_DELIMITER);
  if (search_all_font_in_path(fontname, bold, token, result))
	  return 1;
  while ((token = strtok(NULL, PATH_DELIMITER)))
  {
	  if (search_all_font_in_path(fontname, bold, token, result))
		  return 1;
  }
#endif
  return 0;
}
int get_new_font(char *fontname, int font_size, int bold)
{
	char fontpath[2048] = {0};

	if(!search_for_font(fontname, bold,fontpath))
		if(!search_for_font(fontname, 0,fontpath)) // try without bold setting.
		{
			printf("Failed to find font %s\n",  fontname);
			return 0;
		}


	//printf("fontpath for name is %s %s\n", fontpath, fontname);
	if(font) /* close the font first */
	{
		TTF_CloseFont(font);
		font = NULL;
	}
	  font = TTF_OpenFont(fontpath,font_size);
	  if(!font)
	  {
		  return -1;
	  }

	  if(bold)
		TTF_SetFontStyle(font,TTF_STYLE_BOLD);
	  return 0;
}



int get_font_height()
{
	if(font)
		return TTF_FontHeight(font);
	return -1;
}

static void adjustPlacement(SDL_Rect *rect, int scrwidth, int format)
{
	if(format & JUSTIFY_RIGHT)
		rect->x = scrwidth-rect->w;
	if(format & CENTER )
		rect->x = (scrwidth - rect->w)/2;
}

static void sdl_printf(SDL_Surface *surface, char * text, int format,
					int x, int y, SDL_Color foreground)
{
	char temp [MAX_CHAR_PERLINE];
	int width = 0;
	int height = 0;
	SDL_Rect rect;
	rect.x = x;
	rect.y = y;
	memset(temp,0,MAX_CHAR_PERLINE);

	if(!font)
		get_new_font(FONTNAME,FONTSIZE,0);
  if(!font)
  {
    printf("Font cannot be loaded \n");
    return ;
  }
	TTF_SizeText(font, text, &width, &height);

	if(y+height+marginh > surface->h)
		return ; // skip out of margin.

	if(x+width > surface->w-marginw)
	{
		int stop = 0;
		int stop1 =0;
		int stop2 =0;
		strncpy(temp,text,MAX_CHAR_PERLINE -1);
		{// look for \n or \r
			int i =0;
			for(i =0; i <MAX_CHAR_PERLINE -1; i++)
			{
				if(temp[i] == '\n' || temp[i] == '\r')
				{
					stop1 = i;
					break;
				}
			}
		}
		if(!stop)
		{// no line breaks found. break at a white space
			int i =0;
			for(i =MAX_CHAR_PERLINE-2; i >=0 ; i--)
			{
				int w =0,
					h =0;
				char last = temp[i+1];
				temp[i+1] = 0;
				TTF_SizeText(font, temp,&w, &h);
				if( (temp[i] == ' ' || temp[i] == '\t') &&  w+x < surface->w-marginw )
				{
					stop2 =i;
					break;
				}
				temp[i+1] = last;
			}
		}

		if(stop1 && stop2)
			stop = min(stop1, stop2);
		else if(stop1)
			stop = stop1;
		else if(stop2)
			stop = stop2;
		if(stop)
		{
			SDL_Surface * textsurface = NULL;
			temp[stop] = 0;
			textsurface = TTF_RenderText_Blended(font,temp,foreground);
			if(!textsurface)
			{
				printf("Could not create text surface \n");
				return ;
			}
			rect.w = textsurface->w;
			rect.h = textsurface->h;
			adjustPlacement(&rect, surface->w, format);
			SDL_BlitSurface(textsurface, NULL,surface,&rect);
			if(format & WRAP && strlen(temp) && strlen(temp) < strlen(text))
			{
				int i =stop+1;
				if(stop != stop1) // line break dont skip the spaces
				{
					for(;i < MAX_CHAR_PERLINE -1; i++)
					{
						if(temp[i] != ' ')
							break;
					}
				}
				sdl_printf(surface,text+i,format, x,y+(int)((double)height*(linespacing)),foreground);
			}
			SDL_FreeSurface(textsurface);
			return;
		}
	}
	else
	{
		SDL_Surface * textsurface = NULL;
    if(strlen(text) > MAX_CHAR_PERLINE)
      text[MAX_CHAR_PERLINE-1] = 0; // chop the line.
		strcpy(temp,text);
		textsurface = TTF_RenderText_Blended(font,temp,foreground);
		if(!textsurface)
		{
			printf("Could not create text surface \n");
			return ;
		}
		rect.w = textsurface->w;
		rect.h = textsurface->h;
		adjustPlacement(&rect, surface->w, format);
		SDL_BlitSurface(textsurface, NULL,surface,&rect);
		SDL_FreeSurface(textsurface);
	}
}


/*******************************************************************************
 * sdl_demo_printf
 * Purpose: printf like text printing routine to a surface.
 * Input  : surface
 *			x  - x position on the resource
 *			y  - y position on the resource.
 *          fsize - font size
 *          ftype - font type
 *			text - format string
 *			...  - arguments
 * Output :
 *******************************************************************************/
void graphic_printf(SDL_Surface *surface, SDL_Color fg, int format, int x, int y, const char *fmt, ...)
{
	va_list argptr;
	char *msgbuf = malloc(sizeof(char) *1024*4);


	if(format & WRAP)
	{
		x+= marginx;
		y+= marginy;
	}
	va_start(argptr, fmt);          // Create the text string, similar to printf()
	vsprintf(msgbuf, fmt, argptr);
	va_end(argptr);
	sdl_printf(surface,msgbuf,format, x, y, fg);
	free(msgbuf);
}
